#include <stdio.h>
#include <string.h>

int write_error(void)
{
	fprintf(stderr, "error while writing to output\n");
	return 1;
}

int eof_error(void)
{
	fprintf(stderr, "end of file while reading input\n");
	return 1;
}

int main( int ac, char *av[] )
{
	char filename[255+1], *p;
	unsigned char buffer[32];
	FILE *fi, *fo;
	unsigned offset, datasize, wavesize, riffsize, freq, bits, rate;

	if( ac < 1 )
	{
		fprintf(stderr, "usage: samtowav filename[.sam]\n");
		fprintf(stderr, "Convert MAME sample file to WAV format\n");
		return 1;
	}

	strcpy(filename, av[1]);
	while( (p = strchr(filename, '\\')) != NULL )
		*p = '/';
	p = strrchr(filename, '.');
	if( p && !strchr(p, '/') )
	{
		/* ok, there is an extension */
	}
	else
	{
		strcat(filename, ".SAM");
	}

	fi = fopen(filename, "rb");
	if( !fi )
	{
		fprintf(stderr, "cannot open %s\n", filename);
        return 1;
    }

	p = strrchr(filename, '.');
	strcpy(p, ".WAV");

	fo = fopen(filename, "wb");
	if( !fo )
	{
		fprintf(stderr, "cannot create %s\n", filename);
        return 1;
    }

	offset = fread(buffer, 1, 4, fi);
	if( offset < 4 )
		return eof_error();

    if( strncmp(buffer, "MAME", 4) != 0 )
	{
		fprintf(stderr, "format error: MAME not found\n");
        return 1;
    }
	offset += fread(buffer, 1, 4, fi);
	if( offset < 8 )
		return eof_error();

	datasize = buffer[0] + (buffer[1] << 8) + (buffer[2] << 16) + (buffer[3] << 24);
	wavesize = /* "fmt " */ 4 + 4 + 4 + 2 + 2;
	riffsize = /* RIFF */ 4 + 4 + 4 + 4 + 4 + wavesize + datasize;

    buffer[0] = riffsize & 0xff;
	buffer[1] = (riffsize >> 8) & 0xff;
	buffer[2] = (riffsize >> 16) & 0xff;
	buffer[3] = (riffsize >> 24) & 0xff;

	if( fwrite("RIFF", 1, 4, fo) != 4 )
		return write_error();

    if( fwrite(buffer, 1, 4, fo) != 4 )
		return write_error();

	if( fwrite("WAVE", 1, 4, fo) != 4 )
		return write_error();

	if( fwrite("fmt ", 1, 4, fo) != 4 )
		return write_error();

	buffer[0] = wavesize & 0xff;
	buffer[1] = (wavesize >> 8) & 0xff;
	buffer[2] = (wavesize >> 16) & 0xff;
	buffer[3] = (wavesize >> 24) & 0xff;
	if( fwrite(buffer, 1, 4, fo) != 4 )
        return write_error();

    /* number of channels - alway 1 for MAME samples */
	if( fwrite("\001\000\001\000", 1, 4, fo) != 4 )
		return write_error();

	/* read frequency */
    offset += fread(buffer, 1, 4, fi);
	if( offset < 12 )
        return eof_error();

	freq = buffer[0] + (buffer[1] << 8) + (buffer[2] << 16) + (buffer[3] << 24);

	/* write frequency */
    if( fwrite(buffer, 1, 4, fo) != 4 )
		return write_error();

	/* read bits, amplitude and extension */
    offset += fread(buffer, 1, 4, fi);
    if( offset < 16 )
        return eof_error();

	bits = buffer[0];

	if(bits == 16)
		rate = freq * 2;
	else
		rate = freq;

	buffer[0] = rate & 0xff;
	buffer[1] = (rate >> 8) & 0xff;
	buffer[2] = (rate >> 16) & 0xff;
	buffer[3] = (rate >> 24) & 0xff;

    /* write bytes per second */
    if( fwrite(buffer, 1, 4, fo) != 4 )
		return write_error();

	/* write alignment */
	if( bits == 16 )
	{
		if( fwrite("\002\000", 1, 2, fo) != 2 )
			return write_error();
	}
	else
	{
		if( fwrite("\001\000", 1, 2, fo) != 2 )
			return write_error();
	}

    buffer[0] = bits & 0xff;
	buffer[1] = (bits >> 8) & 0xff;

	/* write bits per sample */
	if( fwrite(buffer, 1, 2, fo) != 2 )
		return write_error();

    if( fwrite("data", 1, 4, fo) != 4 )
		return write_error();

	buffer[0] = datasize & 0xff;
	buffer[1] = (datasize >> 8) & 0xff;
	buffer[2] = (datasize >> 16) & 0xff;
	buffer[3] = (datasize >> 24) & 0xff;
	if( fwrite(buffer, 1, 4, fo) != 4 )
        return write_error();

	if( bits == 16 )
	{
		for( ; ; )
		{
			if( fread(buffer, 1, 2, fi) != 2 )
				break;
			if( fwrite(buffer, 1, 2, fo) != 2 )
				return write_error();
		}
	}
	else
	{
		for( ; ; )
        {
			if( fread(buffer, 1, 1, fi) != 1 )
				break;
			buffer[0] = buffer[0] ^ 0x80;
			if( fwrite(buffer, 1, 1, fo) != 1 )
				return write_error();
        }
    }
	printf("%-12s %dHz, %d bits, %d samples\n",
		filename, freq, bits, datasize * 8 / bits);
    fclose(fo);
	fclose(fi);
	return 0;
}
